local _G = _G
local type = _G.type
local pairs = _G.pairs
local ipairs = _G.ipairs
local select = _G.select
local tonumber = _G.tonumber
local GetTalentInfo = _G.GetTalentInfo
local GetShapeshiftForm = _G.GetShapeshiftForm
local AceLibrary = _G.AceLibrary

local MAJOR_VERSION = "Threat-1.0"
local MINOR_VERSION = tonumber(("$Revision: 37072 $"):match("%d+"))

if _G.ThreatLib_OUT_OF_DATE or not AceLibrary:IsNewVersion(MAJOR_VERSION, MINOR_VERSION) then
	_G.ThreatLib_OUT_OF_DATE = true
	return
end

if MINOR_VERSION > _G.ThreatLib_MINOR_VERSION then
	_G.ThreatLib_MINOR_VERSION = MINOR_VERSION
end

local ThreatLib = _G.ThreatLib

local _, c = _G.UnitClass("player")
if c ~= "DRUID" then return end 

local BS = AceLibrary("Babble-Spell-2.2")
--local Aura = AceLibrary("SpecialEvents-Aura-2.0")
ThreatLib:GetModule("ClassCore"):SetModuleDefaultState(c, false)
local Druid = ThreatLib:GetModule("ClassCore"):NewModule(c)


local HealsEN = { "Healing Touch", "Regrowth", "Rejuvenation", "Lifebloom", "Swiftmend" }
local threatValues = {
	["faerieFire"] = { nil, nil, nil, 108, 127 }, --TODO ranks 1-3
	["maul"] = { nil, nil, nil, nil, nil, nil, 207, 322 }, --TODO ranks 1-6
	["lacerate"] = { 0.2, 285 },
	["mangle"] = 1.3,
	["cower"] = { 240, 390, 600, 800, 1170 },
}

-- recursive function to calculate linear threat for missing spell ranks
local function calcLinearThreat(data)
	if not data then
		return
	end
	if type(data) == "table" then
		for _,v in pairs(data) do
			calcLinearThreat(v)
		end
	elseif type(data) == "string" then
		local tbl = threatValues[data]
		local m = #tbl
		for i=1,m do
			if not tbl[i] then
				tbl[i] = (tbl[m] / m) * i
			end
		end
	end
end
calcLinearThreat({"faerieFire", "maul"})

function Druid:ClassInit()
	self.className = "Druid"
	
	self:GetStanceThreatMod()

	--Cast Handlers
	self.CastHandlers[BS["Growl"]] = self.Growl
	self.CastHandlers[BS["Faerie Fire (Feral)"]] = self.FaerieFire
	self.CastHandlers[BS["Maul"]] = self.Maul
	self.CastHandlers[BS["Lacerate"]] = self.Lacerate
	self.CastHandlers[BS["Cower"]] = self.Cower
	
	--Ability Threat Mods
	
	-- Subtlety for all Heals
	for _,v in ipairs(HealsEN) do
		self.abilityThreatMods[BS[v]] = self.Subtlety
	end
	
	-- Subtlety for all Arcane or Nature Damage, thats all on casts out there
	self.schoolThreatMods["Nature"] = self.Subtlety
	self.schoolThreatMods["Arcane"] = self.Subtlety
	
	self.abilityThreatMods[BS["Tranquility"]] = self.Tranquility
	self.abilityThreatMods[BS["Lacerate"]] = self.LacerateDmg
	self.abilityThreatMods[BS["Mangle (Bear)"]] = self.Mangle
	
	self:RegisterEvent("UPDATE_SHAPESHIFT_FORM", function() self:GetStanceThreatMod() end )
end

function Druid:ScanTalents()
	self.feralinstinctMod = 0.05 * select(5, GetTalentInfo(2, 3))
	self.subtletyMod = 1 - 0.04 * select(5, GetTalentInfo(3, 7))
	self.tranqMod = 1 - 0.5 * select(5, GetTalentInfo(3, 13))
end

-- get form passive threat modifier
-- this function has one issue, if the druid skipped his Aquatic Quest, cat form will have index 2 not 3 - maybe change detection
function Druid:GetStanceThreatMod()
	local form = GetShapeshiftForm()
	if form == 1 then
		self.passiveThreatModifiers = 1.3 + self.feralinstinctMod
	elseif form == 3 then
		self.passiveThreatModifiers = 0.71
	else
		self.passiveThreatModifiers = 1
	end
	self.totalThreatMods = nil		-- Needed to recalc total mods
end

function Druid:ParseRank(rank, max)

	if type(rank) == "string" then
		rank = tonumber(rank:match("%d+"))
	end
	
	if not rank or rank > max then
		rank = max
	end
	
	return rank
end

function Druid:Growl()
	if UnitExists("targettarget") then
		local myThreat = ThreatLib:GetThreat(UnitName("player"), UnitName("target"))
		local theirThreat = ThreatLib:GetThreat(UnitName("targettarget"), UnitName("target"))
		self:SetTargetThreat(max(myThreat, theirThreat))
	else
		-- Assume we get max threat? This seems dangerous, but we need it to apply taunt in situations where the mob is sheeped/stunned/shackled/anything else that prevents a target
		local myThreat = ThreatLib:GetThreat(UnitName("player"), UnitName("target"))
		local theirThreat = ThreatLib:GetMaxThreatOnTarget(UnitName("target"))
		self:SetTargetThreat(max(myThreat, theirThreat))
	end	
end

function Druid:FaerieFire(rank)
	self:AddTargetThreat(threatValues.faerieFire[self:ParseRank(rank, 5)] * self:threatMods())
end

function Druid:Maul(rank)
	self:AddTargetThreat(threatValues.maul[self:ParseRank(rank, 8)] * self:threatMods())
end

function Druid:Lacerate(rank)
	self:AddTargetThreat(threatValues.lacerate[self:ParseRank(rank, 1) + 1] * self:threatMods())
end
function Druid:Cower(rank)
	self:AddTargetThreat(-1 * threatValues.cower[self:ParseRank(rank, 5)] * self:threatMods())
end

function Druid:Subtlety(amount)
	return amount * self.subtletyMod
end

function Druid:Tranquility(amount)
	return self:Subtlety(amount * self.tranqMod)
end

function Druid:LacerateDmg(amount)
	return amount * threatValues.lacerate[1]
end

function Druid:Mangle(amount)
	return amount * threatValues.mangle
end

_G.tinsert(ThreatLib.UpvalueFixers, function(lib) ThreatLib = lib end)
